;******************************************************************
;
; PIC LCD Dual channel Digital Volt Meter building brik.
;
; It uses the Linear Technology LTC1298 dual channel AD converter and
; the Microchip Technology pic 16F84. 
; I use an 2 x 20 characters LCD display but other are suitable.
; Din and Dout of the ADC are connected to RA0 of pic, CLK is connected 
; to RA1 and CS to RA2. RA0 direction is reversed for read/write.
; The DVM reeds 4095 for 5 volt input. The accuracy depends by the 
; accuracy of the 5V power supply. Look at the LTC1286/LTC1298 data sheet.
; RA3, RA4 and RB0 are free for future expansions.
; The program needs a 4 MHz xtal clock. Tune the delays concerning ADC
; read, write and clock for other pic clock frequencies.
;
; by Claudio Pozzi (ik2pii) e-mail <ik2pii@amsat.org>
;
; 24 oct 1999
;
;******************************************************************

	title   "PIC LCD Dual channel Digital Volt Meter (adcpic.asm)"
	list p=pic16f84,r=dec,n=80,x=off,st=off
	include <p16f84.inc>
	errorlevel -302         ; no bank warnings

clock   equ     4000000         ; my crystal frequency
dusec   equ     1000            ; required delay (1 mSec)
	include <timer.inc>     ; (note include must be after 
				;  previous two definitions)

dodigit macro   arg1
	movlw   high(arg1)
	movwf   shi
	movlw   low(arg1)
	movwf   slo
	call    dosub
	endm

clockadc        macro           ; tune for clock frequency
	bsf     PORTA,ADC_CLK
	nop
	nop
	bcf     PORTA,ADC_CLK
	nop
	endm

dotris  macro   arg1, arg2
	bsf     STATUS,RP0
	movlw   arg1
	movwf   arg2
	bcf     STATUS,RP0
	endm

dodelay macro   arg1
	movlw   arg1
	call    delay
	endm
       

ADC_DATA        equ     0       ; RA0
ADC_CLK         equ     1       ; RA1
ADC_CS          equ     2       ; RA2

; Data -
	cblock  0xc             ; start of data area
	dval                    ; delay routine counter
	counter                 ; adc routine bit counter
	temp                    ; temporary register
	hi                      ; 16 bit binary value 8 msb
	lo                      ; 8 lsb
	shi                     ; for bin2ascii routine
	slo
	address                 ; LCD position for string
	ascii:5                 ; ascii string
	eos                     ; end of string =0x00
	endc                    ; end of data area

;****************************************************************

	org     0               ; for future expansion
	goto    start
	org     4
	retfie
	
;****************************************************************
; Main program begins

start   clrwdt                          ; setup 1mSec timer
	bsf     STATUS,RP0
	movlw   b'11010000' | PRESCL    ; prescaler to TMR0
	movwf   OPTION_REG
	bcf     STATUS,RP0
	call    lcd_init                ; initialise the LCD driver
loop                                    ; add some nop for 10 MHz clock
	movlw   b'00000110'             ; CS=1  CLK=1 DATA=0
	movwf   PORTA                   
	dotris  b'00011000',PORTA       ; CS, CLK and DATA = out
	
	movlw   b'00000010'             ; CS=0  CLK=1 DATA=0
	movwf   PORTA                   
	
	movlw   b'00000001'             ; CS=0  CLK=0 DATA=1
	movwf   PORTA                   ; start bit
	movlw   b'00000011'
	movwf   PORTA
	
	movlw   b'00000001'             ; CS=0  CLK=0 DATA=1
	movwf   PORTA                   ; single channel
	movlw   b'00000011'
	movwf   PORTA
	
	movlw   b'00000000'             ; CS=0  CLK=0 DATA=0
	movwf   PORTA                   ; channel = 0
	movlw   b'00000010'
	movwf   PORTA

	call    readadc                 ; read ADC CH0
	call    bin2ascii               ; convert into ASCII string
	movlw   0x00
	movwf   address                 ; LCD row 1, col 1
	movlw   address
	call    sendstr                 ; display value
 
	dotris  b'00011000',PORTA       ; CS, CLK and DATA = out
	movlw   b'00000010'             ; CS=0  CLK=1 DATA=0
	movwf   PORTA                   
	
	movlw   b'00000001'             ; CS=0  CLK=0 DATA=1
	movwf   PORTA                   ; start bit
	movlw   b'00000011'
	movwf   PORTA
	
	movlw   b'00000001'             ; CS=0  CLK=0 DATA=1
	movwf   PORTA                   ; single channel
	movlw   b'00000011'
	movwf   PORTA
	
	movlw   b'00000001'             ; CS=0  CLK=0 DATA=1
	movwf   PORTA                   ; channel = 1
	movlw   b'00000011'
	movwf   PORTA

	call    readadc                 ; read ADC CH1

	call    bin2ascii               ; convert into ASCII string
	movlw   D'10'
	movwf   address                 ; LCD row 1, col 10
	movlw   address
	call    sendstr                 ; display value
	
	dodelay 200
	dodelay 100
	goto    loop

;****************************************************************
;
; readadc: it terminate to read ADC 
;
; INPUT:  MSB, channel and mode programmed in ADC
; OUTPUT: hi,lo => 16 bit binary value
; USES:   counter, temp
; BUGS:   To Be Verified
;
;****************************************************************

readadc
	clrf    lo
	clrf    hi
	clockadc                        ; clock in MSB first
	dotris  b'11111001',PORTA       ; ADC_DATA=in
	clockadc                        ; dummy clock
	movlw   12                      ; 12 bits
	movwf   counter
adcloop
	clockadc
	movf    PORTA,w                 ; read bit ...
	andlw   b'00000001'
	movwf   temp
	rrf     temp,f
	rlf     lo,f                    ; ... and store it
	rlf     hi,f
	decfsz  counter,f
	goto    adcloop
	movlw   b'11111110'
	movwf   PORTA                   ; CS=1  CLK=1 disable ADC
	return

;****************************************************************
;
; bin2ascii: 16 bit unsigned binary to ASCII converter
;
; INPUT:  hi,lo => 16 bit binary value
; OUTPUT: ascii => 5 digit ASCII string, NULL terminated
; USES:   FSR, shi, slo
; BUGS:   ascii string is left padded with "0" (not a real bug but...)
;
;****************************************************************

bin2ascii
	movlw   "0"-1                   ; fill ascii string with ("0"-1)
	movwf   ascii+0
	movwf   ascii+1
	movwf   ascii+2
	movwf   ascii+3
	movwf   ascii+4

	movlw   ascii
	movwf   FSR
	dodigit D'10000'
	dodigit D'1000'
	dodigit D'100'
	dodigit d'10'
	movf    lo,w                    ; convert last digit
	addlw   "0"
	movwf   INDF
	clrf    eos                     ; terminate string with 0x00
	return

dosub
	incf    INDF,f                  ; inc ascii char
	movf    slo,w
	subwf   lo,f
	movf    shi,w
	btfss   STATUS,C
	addlw   1
	subwf   hi,f
	btfsc   STATUS,C
	goto    dosub

	movf    slo,w                   ; reverse the last subcration
	addwf   lo,f
	movf    shi,w
	btfsc   STATUS,C
	addlw   1
	addwf   hi,f
	incf    FSR,f                     ; inc ascii ptr
	return

;****************************************************************
;
; sendstr: it send an ASCII string in data memory to the LCD.
;
; INPUT:  w => pointer to the string, the string must be NULL
;              terminated
; OUTPUT: 
; USES:   FSR
; BUGS:   To Be Verified
;
;****************************************************************

sendstr  
	movwf   FSR                     ; set ptr
	movf    INDF,w
	iorlw   0x80                    ; set as address command
	call    lcd_c                   ; send command
snd1    
	incf    FSR,f                   ; inc ptr
	movf    INDF,w
	iorlw   0                       ; terminating null?
	btfsc   STATUS,Z        
	return                          ; yes, done
	call    lcd_d                   ; send to LCD
	goto    snd1
	
;****************************************************************
;
; Delay routine using the internal timer. Will delay for W
; times the value of "dusec" in microseconds.
;
;****************************************************************

delay   movwf   dval
dy1     clrf    TMR0
	bcf     INTCON,T0IF
	movlw   TMRVAL                  ; set timer
	movwf   TMR0
dy2     btfss   INTCON,T0IF             ; timer overflow yet?
	goto    dy2                     ; no
	decfsz  dval,f                  ; yes, all cycles?
	goto    dy1                     ; no
	return                          ; yes, delay finished
	
	include <lcd.asm>

	end
